#include "windows.h"
#include "MFCStringImpl.h"
using NotMFC::CString;

#include "regkey.h"

CRegKey::CRegKey()
{
	m_hKey = NULL;
}

CRegKey::~CRegKey()
{
	Close();
}

LONG CRegKey::Open (HKEY hKeyRoot, const char* pszPath)
{
	DWORD dw;
	m_sPath = pszPath;

	return RegCreateKeyEx (hKeyRoot, pszPath, 0L, NULL,
		REG_OPTION_VOLATILE, KEY_ALL_ACCESS, NULL, 
		&m_hKey, &dw);
}

void CRegKey::Close()
{
	if (m_hKey)
	{
		RegCloseKey (m_hKey);
		m_hKey = NULL;
	}
}

LONG CRegKey::Write (const char* pszKey, DWORD dwVal)
{
	return RegSetValueEx (m_hKey, pszKey, 0L, REG_DWORD,
		(CONST BYTE*) &dwVal, sizeof(DWORD));
}

LONG CRegKey::Write (const char* pszKey, const char* pszData)
{


	return RegSetValueEx (m_hKey, pszKey, 0L, REG_SZ,
		(CONST BYTE*) pszData, strlen(pszData) + 1);
}

LONG CRegKey::Write (const char* pszKey, const BYTE* pData,
	DWORD dwLength)
{
	return RegSetValueEx (m_hKey, pszKey, 0L, REG_BINARY,
		pData, dwLength);
}

LONG CRegKey::Read (const char* pszKey, DWORD& dwVal)
{

	DWORD dwType;
	DWORD dwSize = sizeof (DWORD);
	DWORD dwDest;

	LONG lRet = RegQueryValueEx (m_hKey, (LPSTR) pszKey, NULL, 
		&dwType, (BYTE *) &dwDest, &dwSize);

	if (lRet == ERROR_SUCCESS)
		dwVal = dwDest;

	return lRet;
}


LONG CRegKey::Read (const char* pszKey, CString& sVal)
{

	DWORD dwType;
	DWORD dwSize = 200;
	char  string[200];

	LONG lReturn = RegQueryValueEx (m_hKey, (LPSTR) pszKey, NULL,
		&dwType, (BYTE *) string, &dwSize);

	if (lReturn == ERROR_SUCCESS)
		sVal = string;

	return lReturn;
}



LONG CRegKey::ReadNext(CString & sVal, int index)
{
	
	DWORD dwSize = 200;
	char  string[200];
	FILETIME ftime;

	return RegEnumKeyEx(m_hKey, index,string, &dwSize,0,0,0,&ftime);
}



LONG CRegKey::Read (const char* pszKey, BYTE* pData, DWORD& dwLen)
{

	DWORD dwType;

	return RegQueryValueEx (m_hKey, (LPSTR) pszKey, NULL,
		&dwType, pData, &dwLen);

}

