#include "sybase.hpp"
#include "applic.h"
#include <stdlib.h>

//		---------		SyConnection implementation		--------------------
//		____________________________________________________________________
LstList SyConnection::list;

SyConnection::SyConnection(CS_CONTEXT* _context)	{
	context = _context;
	connection = 0;
	hasTransaction = FALSE;
	hasError = FALSE;
	db = 0;
	object = 0;
	client = 0;
	}

Result SyConnection::release()	{
	if (hasTransaction)
		if (rollbackTrans() == FAIL)
			return FAIL;
	if (detach() == FAIL)
		return FAIL;
	return free();
	}

Result SyConnection::setLocale(const char* locName)	{
	CS_LOCALE* csl;
	if (cs_loc_alloc(context,&csl) != CS_SUCCEED)
		return FAIL;
	if (cs_locale(context,CS_SET,csl,CS_LC_ALL,(char*)locName,CS_NULLTERM,NULL)
																					 != CS_SUCCEED)
		return FAIL;

	if (ct_con_props(connection, CS_SET, CS_LOC_PROP, csl, CS_UNUSED, NULL)  != CS_SUCCEED)
		return FAIL;

	if (cs_loc_drop(context,csl) != CS_SUCCEED)
		return FAIL;
	return OK;
	}

Result SyConnection::attach(SyConnectionHolder* _db, const char* locName)	{
	db = _db;
	if (context == 0)
		context = db->context;
	if (connection == 0)
		if (alloc() == FAIL)
			return FAIL;
	if (ct_con_props(connection, CS_SET, CS_USERNAME,
						db->userName, CS_NULLTERM, NULL)  != CS_SUCCEED)
		return FAIL;
	if (ct_con_props(connection, CS_SET, CS_PASSWORD,
						db->password, CS_NULLTERM, NULL)  != CS_SUCCEED)
		return FAIL;
	if (locName == 0)
		locName = db->locName;
	if (locName)
		if (setLocale(locName) == FAIL)
			return FAIL;

	if (ct_connect(connection,db->fileName,CS_NULLTERM) != CS_SUCCEED)
		return FAIL;

	int textLimit = MAXLONG;
	if (ct_con_props(connection, CS_SET, CS_TEXTLIMIT,&textLimit,
						CS_UNUSED, NULL)  != CS_SUCCEED)
		return FAIL;

	if (ct_options(connection, CS_SET, CS_OPT_TEXTSIZE, &textLimit,
						CS_UNUSED, NULL)  != CS_SUCCEED)
		return FAIL;

	hasError = FALSE;
	list.insertAtTail(this);		//		Make it active

	SyQuery q;
	if (q.staticExecute("set transaction isolation level 3") == FAIL)
		return FAIL;
	if (db->database[0])
		if (q.unprepare() == FAIL ||
				q.staticExecute(StdFormat("use %s",db->database)) == FAIL)
			return FAIL;
	return OK;
	}

Result SyConnection::detach()	{
	if (hasTransaction)
		return queryError(QUE_DISCONNECT_TRANS);
	if (list.isContains(this))
		list.detach(this);
	if (connection == 0)
		return OK;
	if (ct_close(connection,CS_FORCE_CLOSE) == CS_FAIL)
		return FAIL;
	return OK;
	}

Result SyConnection::alloc()	{
	if (ct_con_alloc(context,&connection) != CS_SUCCEED)
		return FAIL;
	return OK;
	}

Result SyConnection::free()	{
	if (connection == 0)
		return OK;
	if (ct_con_drop(connection) != CS_SUCCEED)
		return FAIL;
	connection = 0;
	return OK;
	}

Result SyConnection::activate()	{
	if (!list.isContains(this))
		return queryError(QUE_BAD_ACTIVATE);
	if (list.getLast() != this)	{
		list.detach(this);
		list.insertAtTail(this);
		}
	return OK;
	}

int SyConnection::count(SyConnectionHolder* holder)	{
	int n = 0;
	SyConnection* conn;
	for (LstIterForward(list,conn))
		if (conn->db == holder)	n++;
	return n;
	}

SyConnection* SyConnection::findFree(SyConnectionHolder* db)	{
	SyConnection* conn;
	for (LstIterForward(list,conn))
		if (conn->db == db && !conn->hasTransaction)
			return conn;
	return 0;
	}

void SyConnection::deActivate()	{
	if (list.isContains(this) && list.getFirst() != this)	{
		list.detach(this);
		list.insertAtHead(this);
		}
	}

Result SyConnection::poll(SyConnectionHolder* db)	{
	Result result = OK;
	Pointer conn, next;
	for (conn = list.getFirst(); conn; conn = next)	{
		next = list.getNext(conn);
		SyConnection* c = (SyConnection*)conn;
		if (c->db == db && c->hasTransaction)
			result = queryError(QUE_DISCONNECT_TRANS);

		if (c->db == db || !c->hasTransaction && count(c->db) > c->db->optimalSize)	{
			list.detach(c);
			if (c->release() == FAIL)	{
				list.insertBefore(c,next);		//		Return it back
				result = FAIL;
				}
			else
				if (c->object)
					delete c;
			}
		}
  	return result;
	}

//	----		Transaction control functions
Result SyConnection::startTrans()	{
	if (activate() == FAIL)
		return FAIL;
	if (hasTransaction)
		return queryError(QUE_START_TWICE);

	SyQuery q;
	if (q.staticExecute("BEGIN TRANSACTION") == FAIL)	{
		deActivate();
		return FAIL;
		}
	hasTransaction = TRUE;
	return OK;
	}

Result SyConnection::commitTrans()	{
	if (!hasTransaction)
		return queryError(QUE_COMMIT_NO_TRANS);

	if (activate() == FAIL)
		return FAIL;

	SyQuery q;
	if (q.staticExecute("COMMIT WORK") == FAIL)
		return FAIL;
	deActivate();
	hasTransaction = FALSE;
	return OK;
	}

Result SyConnection::rollbackTrans()	{
	if (!hasTransaction)
		return queryError(QUE_ROLLBACK_NO_TRANS);

	if (activate() == FAIL)
		return FAIL;

	if (client)	{
//		if (ct_cancel(NULL,client->cmd,CS_CANCEL_CURRENT) != CS_SUCCEED
//				||	client->send() != CS_SUCCEED
//				|| client->handleResults() == FAIL)
		if (ct_cancel(connection,NULL,CS_CANCEL_ALL) != CS_SUCCEED)
			return FAIL;
		client = 0;
		}
	SyQuery q;
	Result result = q.staticExecute("ROLLBACK WORK");
	hasTransaction = FALSE;
	deActivate();
	return result;
	}

//	----		Error control functions
void SyConnection::message(int severity, char* text)	{
	SyConnection* conn = SyConnection::getActive();
	if (severity == 10)
		theApp->print(text,TRUE);
	else	{
		if (conn)	{
			conn->hasError = TRUE;
			theApp->print(text,FALSE);			//		for a while
			queryError(QUE_SERVER_ERROR,text);
			}
		else	if (severity < 17)
				theApp->print(text,FALSE);
		if (severity >= 17)	{
			theApp->error();
			theApp->print("Can not recover, stopping...",FALSE);
			theApp->exit(FAIL);
			}
		}
	}

void SyConnection::clearError()	{
	SyConnection* conn = SyConnection::getActive();
	if (conn)
		conn->hasError = FALSE;
	}

Bool SyConnection::isError()	{
	SyConnection* conn = SyConnection::getActive();
	return (conn && conn->hasError);
	}
//		____________________________________________________________________

//	--------		Context implementation (including callbacks)		----------
//	______________________________________________________________________
/*
#include "error.h"
static char* syErrorMessages[] = { "%s", 0 };

class SyError : public ErrorClass	{
public :
	SyError() : ErrorClass("SyError",syErrorMessages)	{}
	virtual void makeMessage(ErrorIndex errorIndex, va_list ap);
	};
*/

CS_RETCODE CS_PUBLIC ex_clientmsg_cb(CS_CONTEXT*,CS_CONNECTION*, CS_CLIENTMSG* errmsg)	{
	MemBuf errText(MemBuf::TEXT_MODE);
	errText << StdFormat(
			"Open Client Message: LAYER=%ld ORIGIN=%ld SEVERITY=%ld NUMBER=%ld\n",
			CS_LAYER(errmsg->msgnumber), CS_ORIGIN(errmsg->msgnumber),
			CS_SEVERITY(errmsg->msgnumber), CS_NUMBER(errmsg->msgnumber));
	errText << errmsg->msgstring;
	if (errmsg->osstringlen > 0)
		errText << "\nOperating System Error: " << errmsg->osstring << "\n";
	SyConnection::message(CS_SEVERITY(errmsg->msgnumber),errText.getBuf());
	return CS_SUCCEED;
	}

CS_RETCODE CS_PUBLIC ex_servermsg_cb(CS_CONTEXT*,CS_CONNECTION*, CS_SERVERMSG* srvmsg)	{
	MemBuf errText(MemBuf::TEXT_MODE);
	errText << StdFormat("Server message: %ld, Severity %ld, State %ld, Line %ld",
				srvmsg->msgnumber, srvmsg->severity, srvmsg->state, srvmsg->line);
	if (srvmsg->svrnlen > 0)
		errText << " (Server " << srvmsg->svrname << ")";
	if (srvmsg->proclen > 0)
		errText << " (Procedure " << srvmsg->proc << ")";
	errText << "\n" << srvmsg->text;
	SyConnection::message(srvmsg->severity,errText.getBuf());
	return CS_SUCCEED;
	}

static CS_CONTEXT* context_init()	{
	CS_CONTEXT* cs;
	if (cs_ctx_alloc(CS_VERSION_100,&cs) == CS_SUCCEED)
		if (ct_init(cs,CS_VERSION_100) == CS_SUCCEED)
			if (ct_callback(cs, NULL, CS_SET, CS_CLIENTMSG_CB,
						(CS_VOID*)ex_clientmsg_cb) == CS_SUCCEED &&
				ct_callback(cs, NULL, CS_SET, CS_SERVERMSG_CB,
						(CS_VOID *)ex_servermsg_cb) == CS_SUCCEED)
				return cs;
	return 0;
	}
CS_CONTEXT* theContext = 0;
//	______________________________________________________________________


//	-----------		SyConnectionHolder implementation		----------------
//	______________________________________________________________________
LstList SyConnectionHolder::holders;

SyConnectionHolder::SyConnectionHolder(short size, CS_CONTEXT* _context,
									const char *_fileName, const char* _userName,
									const char* _password, const char* _database)	{
	init(size,_context,_fileName,_userName,_password,_database);
	}

void SyConnectionHolder::init(short size, CS_CONTEXT* _context,
									const char *_fileName, const char* _userName,
									const char* _password, const char* _database)	{
	optimalSize = size;
	if (_context == 0)
		_context = theContext;
	if (_context == 0)
		_context = theContext = context_init();
	context = _context;
	fileName = strdup(_fileName ? _fileName : "");
	userName = strdup(_userName ? _userName : "");
	password = strdup(_password ? _password : "");
	database = strdup(_database ? _database : "");
	locName = 0;
	holders.insertAtHead(this);
	}

void SyConnectionHolder::setLocale(const char* loc)	{
	if (locName)	free(locName);
	if (loc)
		locName = strdup(loc);
	else
		locName = 0;
	}


SyConnectionHolder::~SyConnectionHolder()	{	drop();	}

Result SyConnectionHolder::drop()	{
	if (SyConnection::poll(this) == FAIL)
		return FAIL;
	if (holders.isContains(this))
		holders.detach(this);
	free(fileName);
	free(userName);
	free(password);
	free(database);
	free(locName);
	return OK;
	}
//		_____________________________________________________________________


//		---------------		SyDatabase implementation		---------------------
//		_____________________________________________________________________
Result SyDatabase::ready(const char *fileName, const char* userName,
																const char* password)	{
	char buff[MAXSTR];
	strcpy(buff,fileName);
	char* server = strtok((char*)fileName,":");
	char* database = strtok(0,":");
	if (database == 0)
		database = "master";
	if (theContext == 0)
		if ((theContext = context_init()) == 0)
			return FAIL;
	init(2,0,server,userName,password,database);
	return OK;
	}

Result SyDatabase::finish ()  {
//	StackOfQuery::flush();
	return OK;
	}
//		_____________________________________________________________________

//		--------------		CsCommand implementation		---------------------
//		_____________________________________________________________________
Result CsCommand::alloc()	{
	if (cmd)
		return OK;
	trans = SyConnection::getActive();
	if (ct_cmd_alloc(trans->connection,&cmd) != CS_SUCCEED)
		return FAIL;
	return OK;
	}

Result CsCommand::free()	{
	if (cmd == 0)
		return OK;
	if (ct_cmd_drop(cmd) != CS_SUCCEED)
		return FAIL;
	cmd = 0;
	return OK;
	}

CS_RETCODE CsCommand::send()	{
	trans->hasError = FALSE;
	CS_RETCODE res = ct_send(cmd);
	if (res == CS_SUCCEED)
		trans->client = this;
	return res;
	}

Result CsCommand::handleResults()	{
	CS_INT res_type;
	CS_RETCODE retcode;
	while ((retcode = getResult(&res_type)) == CS_SUCCEED)
		if (handleResult(res_type) == CS_FAIL)
			return FAIL;
	return  (retcode == CS_END_RESULTS ? OK : FAIL);
	}

CS_RETCODE CsCommand::handleResult(CS_INT res_type)	{
	switch ((int)res_type)	{
		case CS_CMD_SUCCEED:	// Command returning no rows completed successfully.
		case CS_CMD_DONE:		//	This means we're done with one result set.
			break;
		case CS_CMD_FAIL:		//	This means that the server encountered
			return CS_FAIL; 	//	an error while processing our command.

		default:						//		We got an unexpected result type
			theApp->print(StdFormat("res_type = %d",(int)res_type),FALSE);
			return CS_FAIL;
		}
	return CS_SUCCEED;
	}

CS_RETCODE CsCommand::getResult(CS_INT* res_type)	{
	if (SyConnection::isError())	{
		SyConnection::clearError();
		return CS_FAIL;
		}
	CS_RETCODE retcode = ct_results(cmd,res_type);
	switch (retcode)	{
		case CS_SUCCEED :
			break;
		case CS_END_RESULTS:	// Everything went fine
		case CS_FAIL:			//	Something went wrong
			trans->client = 0;
			break;

		default:
			theApp->print(StdFormat("retcode = %d",(int)retcode),FALSE);
			return CS_FAIL;
		}
	return retcode;
	}
//		_____________________________________________________________________

Result SyConnection::connect()	{
	SyConnection* conn = SyConnection::findFree(SyConnectionHolder::currentDb());
	if (conn == 0)	{
		conn = new SyConnection;
		if (conn->attach(SyConnectionHolder::currentDb()) == FAIL)	{
			delete conn;
			conn = 0;
			return FAIL;
			}
		}
	if (conn->activate() == FAIL)
		return FAIL;

	conn->bind((void*)1);			//		This ensures that <conn> will be deallocated
	return OK;
	}

Result SyConnection::disconnect(Bool andDetach)	{
	SyConnection* conn = getActive();
	if (conn == 0)
		return OK;
	conn->deActivate();
	if (andDetach)
		return conn->detach();
	SyConnection::poll();
	return OK;
	}

