#ifndef _SYBASE_H
#define _SYBASE_H
#include	"__query.hpp"
#undef DB_DATE
#include	"date.h"
#include	"dump.hpp"
#include	<ctpublic.h>

#define MAX_QUERY_LENGTH	4096

class SyConnection;
class CsCommand	{
public :
	CS_COMMAND* cmd;
	SyConnection* trans;
	CsCommand()	{	clear();	}

	operator CS_COMMAND* ()	{	return cmd;	}
	Result alloc();
	Result free();
	Result handleResults();
	void clear()	{	cmd = 0;	trans = 0;	}

	CS_RETCODE send();
	CS_RETCODE handleResult(CS_INT res_type);
	CS_RETCODE getResult(CS_INT* res_type);
	};

class SyDate : public CS_DATETIME	{
public :
	SyDate& operator=(const Date& date);
	operator Date();
	static Date base;
	};

//		----	 SyConnectionHolder manages a group of connections to server	--
//		_______________________________________________________________________
class SyConnectionHolder : public LstLink0, public LstList	{
public :
	short optimalSize;
	char* fileName;
	char* userName;
	char* password;
	char* database;
	char* locName;
	CS_CONTEXT* context;

	SyConnectionHolder()	{	context = 0;	}
	SyConnectionHolder(short size, CS_CONTEXT* _context, const char *fileName,
							const char* userName = 0, const char* password = 0,
							const char* database = 0);
	~SyConnectionHolder();

	void init(short size, CS_CONTEXT* _context, const char *_fileName,
				const char* _userName, const char* _password,
				const char* _database);
	Result drop();
	void setLocale(const char* loc);


	static LstList holders;
	static SyConnectionHolder* currentDb()
		{	return (SyConnectionHolder*)holders.getLast();	}
	};

class SyDatabase : public SyConnectionHolder	{
public :
	SyDatabase(void* p = 0)	{}
	Bool isReady()	{	return context != 0;	}
	Result ready(const char *fileName, const char* userName,	const char* password);
	Result finish();
	};


//		-----			SyConnection manages a single connection to DB		--------
//		_______________________________________________________________________
class SyConnection : public LstLink0	{
public :
	SyConnectionHolder* db;
	CS_CONTEXT* context;
	CS_CONNECTION* connection;

	Bool hasTransaction;
	int hasError;
	void* object;
	CsCommand* client;

	SyConnection(CS_CONTEXT* _context = 0);
	~SyConnection()	{	release();	}

	//		Connection functions
	Result setLocale(const char* locName);
	Result alloc();
	Result free();
	Result attach(SyConnectionHolder* db, const char* locName = 0);
	Result detach();
	Result release();
	Result activate();
	void deActivate();
	Bool isOpen()	{	return list.isContains(this);	}
	void bind(void* _object)	{	object = _object;	}

	//		Transaction functions
	Result startTrans();
	Result commitTrans();
	Result rollbackTrans();

	static LstList list;			//		Current element is always last
	static SyConnection* getActive()	{	return (SyConnection*)list.getLast();	}
	static int count(SyConnectionHolder* holder);
	static SyConnection* findFree(SyConnectionHolder* db);
	static Result poll(SyConnectionHolder* db = 0);
	static Result connect();
	static Result disconnect(Bool andDetach = FALSE);

	//		Error control functions
	static void message(int severity, char* text);
	static void clearError();
	static Bool isError();
	};
//		_______________________________________________________________________

class SyBlobDesc	{
public :
	short sqltype;
	CS_INT column;
	CS_IODESC desc;
	SyBlobDesc()	{	column = -1;	}
	};

class SyQuery;
class SyBlob : public LstLink0, public SyBlobDesc, public Dump {
public :
	SyBlob& operator=(const SyBlob& src);
	Result read(SyQuery* owner, CS_INT index, Bool dontLoad = FALSE);
	Result write();
	Result play(Dump& data, char* queryText, ...);
	Result replay(Dump& data);
	};

class SyBlobId	{
public:
	short isNotNull; 	//		It's used because there exists practice to replace
							//		= NO_BLOB with memset(...,0,...);
	short isNotModified;
	SyBlob blob;
	SyBlobId()	{	isNotNull = 0;	isNotModified = 1;	}
	SyBlobId& operator=(const SyBlobId& id);
	int operator==(const SyBlobId& id);
	int operator!=(const SyBlobId& id)	{	return !(*this == id);	}
	void get(MemBuf& buf);
	void get(Dump& buf);
	void set(const MemBuf& buf);
	void set(const Dump& buf);
	Result update();
	};


#define MAX_SYQUERY_NAME	10

#define SYM_EXECUTE	 	0
#define SYM_OPEN_CURSOR 1
#define SYM_STATIC		2

#define	SQF_STATIC		0x02
#define	SQF_RPC			0x04
#define  SQF_UPPERCASE	0x08

class SyQuery : public __Query  {
public :
	CsCommand cmd;
	LstList blobs;
	int flags;
	int bindings;
	int nOutput;
	char cursorName[MAX_SYQUERY_NAME+1];
	virtual char* getCursorName();

	Result bind();
//	Result allocateSQLDA(DbSQLDA& s);
	SyBlob* findBlob(CS_INT column);

	CS_RETCODE handleResult(CS_INT res_type);			//		Only one
	Result handleResults();									//		All in the set

	SyQuery(const char *name = 0);
	virtual ~SyQuery();
	virtual char* getStackName() const;
	void setType(int flags);

	Result get(char* queryText, ...);
	Result executeImmediate(char *queryText)	{	return OK;	}
	Result staticExecute(char *queryText);
	Result SyQuery::staticExecute1(char *text, SQLDA* _in_sqlda);
	Result SyQuery::staticExecute2(char *text, char* data);

	virtual Result close(Bool ignoreError = FALSE);
	virtual Result drop(Bool ignoreError = FALSE);

	virtual Result release();
	virtual void cloneFrom(const __Query& query);
	virtual __Query* dup() const	{	return new SyQuery;	}

	virtual Result openCursor();
	virtual Result _prepare(char* queryText);
	virtual Result _unprepare();
	virtual Result _execute(SQLDA* _in_sqlda = 0, short method = 0);
	virtual Result _open(SQLDA* _in_sqlda = 0, short method = 0);
	virtual Bool _fetch();

	Result RPC(int nOut, char* queryText, ...);
	Bool fetchOutput();
	Bool fetchStatus();

	virtual char* getTextField(int fldNo);
	Bool isBlob(int fldNo);

	//	virtual Bool isChar(int fldNo);
	Result _close(Bool deallocate = FALSE);
	Result passArg(SQLDA* sqlda);

	static short notNullind;

	DbSQLDA desc;
	SQLDA* describeInput();
	Result describeIO(DbSQLDA& s);	//		Gets column description in <s>
	Result bind(DbSQLDA& s);			//		Binds query output described by <s>
												//		with <out> member
	int binded;

	virtual Result setArg(MemBuf& dest, short type, SQLVAR* var, va_list& ap);
	virtual Result getArg(DbSQLDA& sqlda, void* data, short type, int i);

	virtual Result getUniArg(DbSQLDA& sqlda, int i, UniVar* uv);
	virtual Result setUniArg(DbSQLDA& sqlda, int i, UniVar* uv);

	static char semiByteToChar(char b);
	static char charToSemiByte(char c);
	static void binaryToText(MemBuf& dest, int size, char* data);
	static void textToBinary(Dump& dest, char* text);
	};
#endif
