//		--------			Some rules to use Queries for Sybase	-------------
//		1. Textes.
//			There isn't allowed varchars and chars longer than 255, so any
//			longer one have to be text. Text blobs becomes texts and binary ones
//			- images. Also there are some features in text/image handling.
//			--------						Reading.		---------------------------
//			SyQuery allows two ways to read text/image :
//			a)	read through binding (member bindings = TRUE). In this case
//				field will be read as any usual field in any of structures like
//				FString,	Dump, Quark, MemBuf etc. Maximal text that can be read
//				such way is 32K-1
//			b)	separate reading (member bindings = TRUE). The field have to be
//				read until the next fetch. In this case all text/image fields
//				must be last in field list of SELECT query. Class <SyBlobId>
//				allows to read the fields using this method by format "%Q"
//			These methods are applied to all text/image fields in query, and
//			it's impossibly to vary method from field to field.

//			--------						Updating.		-------------------------
//			To update text/image field it's necessary to know its "pointer",
//			which can be obtained while separate reading. (It's storing
//			in the <SyBlobId>. So one have to apply separate reading method
//			if it would be needed to update this field later

//			--------						Inserting.		-------------------------
//			To insert row with text/image fields there are three possibilities :
//			a) Create query, containing text/image fields values
//			b) insert a row without text/image fields, update this row
//				explicitly setting these fields to NULL, (it's need
//				to know key for this row), read this row and update text/image
//				fields using SyBlob->write() (or SyBlobId->update())
//			c) The same as b) except you can specify a some values for text/image
//				field while inserting ('' for instance) so you haven't do update
//				with NULL.
//
//		2. Cursor names.
//			Limited by 10 symbols
//
//		3. Binary.
//			Type %B in query implies now pointer to Dump instead of short+void*.
//			Maximal length of binary read such way is 16K-1
//
//		4. Stored procedures.
//			To call a stored procedure one has to create unnamed SyQuery
//			and call RPC(). First parameter if an amount of parameters
//			that procedure returns, second one is procedure name followed
//			by type symbols (for input and output parameters), then follows
//			arguments and any values (or pointer to variables) according to
//			return parameters. Note that if one provides pointers to variables
//			into which he wants to put return parameters, it won't work. He
//			still needs to process these parameters through fetch & read.
//			There may be three sets of data to return : data returned by
//			SQL statements contained in the procedure (accessed by fetch()),
//			status result - single row contained integer (accessed by
//			fetchStatus()) and return parameters (accessed by fetchOutput()).
//			They are in single row too

#include "__queryf.hpp"
#include "sybase.hpp"
#include "applic.h"
#include "dynarray.h"
#include "codepg.h"
#include "date.h"
#include "fstring.h"
#include "dump.hpp"
#include "timer.h"
#include "membuf.h"
#include <stdio.h>

//		--------------		Sybase specific utilities		------------------------
//		________________________________________________________________________
short SyQuery::notNullind = 0;
Date SyDate::base(1,1,2000);

static char* wfrom[] = {	"GET", "SET", "PUT", "FETCH", "REMOVE",
									"INSERT", "SELECT", "UPDATE", "DELETE", 0 };

static char* wto = "suisd"
						 "isud";

static char* vowels = "aeiouyAEIOUY";

void packName(char* dest, const char* src, int maxlen)	{
	if (strlen(src) <= maxlen)
		strcpy(dest,src);
	else	{
		FStringLong s;
		char buffer[MAXSTR];
		char mask[MAXSTR];
		strcpy(buffer,src);
		strupr(buffer);
		memset(mask,0,sizeof(mask));

		//		Eliminate long standard words
		int offset = 0;
		for (int i = 0; wfrom[i]; i++)	{
			char* w = strstr(buffer+offset,wfrom[i]);
			if (w == 0)	continue;
			int to = (int)(w - buffer);
			for (int j = offset; j < to; j++)
				s += buffer[j];
			mask[s.length()] = 1;
			s += wto[i];
			offset = to + strlen(wfrom[i]);
			}
		s += &buffer[offset];

		int remain = s.length() - maxlen;
		if (remain <= 0)
			strcpy(dest,s);
		else	{
			//		Omit vowels
			strcpy(buffer,s);
			s = "";
			for (i = 0; buffer[i]; i++)
				if (strchr(vowels,buffer[i]) == 0)
					s += buffer[i];
				else	if (mask[i])
					s += buffer[i];
				else	if (remain <= 0)
					s += buffer[i];
				else
					remain--;
			if (remain > 0)					//		Truncate
				s[maxlen] = 0;
			strcpy(dest,s);
			}
		}
	}

#ifndef DB_NO_OLD_SQLDA
#define NTYPES 10
static short _uTypes[] = { SQL_SHORT, SQL_LONG,
									SQL_FLOAT, SQL_DOUBLE,
									SQL_FIXCHAR, SQL_VARCHAR,
									SQL_DATE, SQL_QUAD,
									SQL_LTEXT, SQL_BLOB	};

static short _sTypes[] = { CS_SMALLINT_TYPE, CS_INT_TYPE,
									CS_REAL_TYPE, CS_FLOAT_TYPE,
									CS_CHAR_TYPE, CS_CHAR_TYPE,
									CS_DATETIME_TYPE, CS_ILLEGAL_TYPE,
									CS_TEXT_TYPE, CS_IMAGE_TYPE };

short typeToSybase(short type)	{
	for (int i = 0; i < NTYPES; i++)
		if (_uTypes[i] == type)
			return _sTypes[i];
	return type;
	}

short typeFromSybase(short type)	{
	for (int i = 0; i < NTYPES; i++)
		if (_sTypes[i] == type)
			return _uTypes[i];
	return type;
	}

static void SQLimport(SQLVAR& var,const CS_DATAFMT& df)	{
	var.sqltype = typeFromSybase(df.datatype);
	var.sqllen = df.maxlength;
	var.sqlname_length = min(df.namelen,sizeof(var.sqlname)-1);
	memcpy(var.sqlname,df.name,var.sqlname_length);
	var.sqlname[var.sqlname_length] = 0;
	}

static void SQLexport(SQLVAR& var, CS_DATAFMT& df)	{
	df.datatype = typeToSybase(var.sqltype);
	df.maxlength = var.sqllen;
	df.namelen = min(var.sqlname_length,sizeof(df.name)-1);
	memcpy(df.name,var.sqlname,df.namelen);
	df.name[df.namelen] = 0;
	df.scale = 0;
	df.precision = 0;
	}
#endif

SyDate::operator Date()	{
	CS_DATEREC daterec;
	if (cs_dt_crack(SyConnectionHolder::currentDb()->context,CS_DATETIME_TYPE,this,&daterec) == CS_FAIL)
		return Date::errorDate;
	if (daterec.dateyear < Date::minYearNo)
		return Date::minDate;
	if (daterec.dateyear > Date::maxYearNo)
		return Date::maxDate;
	return Date(daterec.datedmonth,daterec.datemonth+1,daterec.dateyear,
					daterec.datehour, daterec.dateminute, daterec.datesecond);
	}

long _div(long a, long b)	{
	if (a >= 0)
		return a/b;
	if ((-a)%b == 0)
		return a/b;
	return -1-(-a)/b;
	}

long _mod(long a, long b)	{
	return a - b*_div(a,b);
	}


SyDate& SyDate::operator=(const Date& date)	{
	long delta = date - base;
	//		Operators "/" and "%" doesn't work well with negative values
	dtdays = 25*(365*4+1) - 1 + _div(delta,SEC_PER_DAY);
	dttime = _mod(delta,SEC_PER_DAY)*300;
	return *this;
	}
//		________________________________________________________________________



//	___________________________________________________________________

//	----------------			SyQuery implementation		----------------
//	___________________________________________________________________
//	___________________________________________________________________

//	----------------					Common methods			----------------
//	___________________________________________________________________
//		Only valuable part of SQLDA will be copied (only sqld SQLVARS),
//		and sqln will become equal sqld
void SyQuery::cloneFrom(const __Query& query)	{
	__Query::cloneFrom(query);
	SyQuery& q = (SyQuery&)query;
	cmd = q.cmd;
	state = DQS_PREPARED | DQS_CACHED | DQS_HAS_STMT;
	if (q.state & DQS_HAS_CURSOR)
		state |= DQS_HAS_CURSOR;
	binded = 0;
	bindings = q.bindings;
	flags = q.flags;
	}

char* SyQuery::getCursorName()	{
	sprintf(cursorName,"_%p",cmd.cmd);
	return cursorName;
	}

SyQuery::~SyQuery  ()  {
	release();
	LstFlush(blobs,SyBlob);
	}

SyQuery::SyQuery(const char *name) : __Query(name)	{
	bindings = TRUE;
	binded = 0;
	flags = SQF_UPPERCASE;
	};

void SyQuery::setType(int _flags)	{
	if (_flags & SQF_STATIC)
		state &= ~DQS_CACHED;
	flags = _flags;
	}

char* SyQuery::getStackName() const	{
	SyConnection* conn = SyConnection::getActive();
	if (conn == 0 || conn->connection == 0)
		return __Query::getStackName();
	return StdFormat("_%p",conn->connection);
	}
//	___________________________________________________________________


//	----------------		Query handling methods			----------------
//	___________________________________________________________________
Result SyQuery::describeIO(DbSQLDA& s)	{
	//		Get description
	CS_INT sqld, sqln, size;
	CS_DATAFMT df;
	if (ct_res_info(cmd,CS_NUMDATA,&sqld,CS_UNUSED,&size) != CS_SUCCEED)
		return FAIL;
	sqln = max(sqld,1);

	s.clear();
	s.hasRecord = FALSE;
	s.extend(sqln);
	((SQLDA*)s)->sqld = sqld;

	for (CS_INT i = 1; i <= sqld; i++)	{
		if (ct_describe(cmd,i,&df) != CS_SUCCEED)
			return FAIL;
//		if ((df.datatype == CS_TEXT_TYPE || df.datatype == CS_IMAGE_TYPE)
//																&& format[i-1] == Ft_Blob_)
//		if ((df.datatype == CS_TEXT_TYPE || df.datatype == CS_IMAGE_TYPE))
//			df.maxlength = 0;
//		SQLimport(sqlda->sqlvar[i-1],df);
		SQLVAR var;
		SQLimport(var,df);
		var.sqldata = 0;
		var.sqlind = 0;
		if (var.sqllen > MAXSHORT)
			if (bindings)
				var.sqllen = MAXSHORT;
			else
				var.sqllen = 0;
		s[i-1] = var;
		}
	return OK;
	}

Result SyQuery::passArg(SQLDA* sqlda)	{
	if (sqlda == NULL)
		return OK;
	for (CS_INT i = 0; i < sqlda->sqld; i++)	{
		SQLVAR* var = &sqlda->sqlvar[i];
		CS_DATAFMT df;
		memset(&df,0,sizeof(df));
		if ((flags & SQF_RPC) && i >= sqlda->sqld - nOutput)
			df.status = CS_RETURN;
		else
			df.status = CS_INPUTVALUE;
		switch ((var->sqltype & ~1))	{
			case SQL_SHORT :	df.datatype = CS_SMALLINT_TYPE;		break;
			case SQL_LONG :	df.datatype = CS_INT_TYPE;				break;
			case SQL_FLOAT :  df.datatype = CS_REAL_TYPE;  			break;
			case SQL_DOUBLE : df.datatype = CS_FLOAT_TYPE; 			break;
			case SQL_DATE :	df.datatype = CS_DATETIME_TYPE; 	 	break;
			default :
				df.datatype = CS_CHAR_TYPE; 			break;
			}
		if (var->sqlind == 0 || *var->sqlind != nullind)
			var->sqlind = &notNullind;
		if (df.datatype == CS_CHAR_TYPE && !*var->sqlind && var->sqllen == 0)	{
			if (ct_param(cmd,&df," ",1,*var->sqlind) == CS_FAIL)
				return FAIL;
			}
		else
			if (ct_param(cmd,&df,var->sqldata,var->sqllen,*var->sqlind) == CS_FAIL)
				return FAIL;
		}
	return OK;
	}

Result SyQuery::bind(DbSQLDA& s)	{
	out.clear();
	out.create(s,TRUE);
	out_sqlda = (SQLDA*)out;
	for (CS_INT i = 0; i < out_sqlda->sqld; i++)	{
		if (!bindings && isBlob(i))
			break;												//		Stop binding
		CS_DATAFMT df;
		memset(&df,0,sizeof(df));
		SQLexport(out_sqlda->sqlvar[i],df);
		df.count = 1;
		if (df.datatype == CS_IMAGE_TYPE)	{
			df.datatype = CS_TEXT_TYPE;
			df.maxlength /= 2;
			}
		if (df.datatype == CS_CHAR_TYPE || df.datatype == CS_TEXT_TYPE)	{
			df.maxlength++;
			df.format = CS_FMT_NULLTERM;
			}
		else
			df.format = CS_FMT_UNUSED;
		if (ct_bind(cmd,i+1,&df,out_sqlda->sqlvar[i].sqldata,0,
								out_sqlda->sqlvar[i].sqlind) != CS_SUCCEED)
			return FAIL;
		}
	binded = 1;
	return OK;
	}

CS_RETCODE SyQuery::handleResult(CS_INT res_type)	{
	switch ((int)res_type)	{
		case CS_ROW_RESULT:	//		Result contains some rows
		case CS_CURSOR_RESULT:	//		Result contains some rows
			if (describeIO(desc) == FAIL)
				return CS_FAIL;
			if (bind(desc) == FAIL)
				return FAIL;
			binded = 1;
			break;

		case CS_PARAM_RESULT :	//		Single row
			if (describeIO(desc) == FAIL)
				return CS_FAIL;
			if (bind(desc) == FAIL)
				return FAIL;
			binded = 2;
			break;

		case CS_STATUS_RESULT :
			if (describeIO(desc) == FAIL)
				return CS_FAIL;
			if (bind(desc) == FAIL)
				return FAIL;
			binded = 3;
			break;

	//	case CS_COMPUTE_RESULT:
		case CS_DESCRIBE_RESULT :
			if (describeIO(desc) == FAIL)
				return CS_FAIL;
			break;

		default :
			return cmd.handleResult(res_type);
		}
	return CS_SUCCEED;
	}

Bool SyQuery::fetchOutput()	{
	return binded == 2 ? _fetch() : FALSE;
	}

Bool SyQuery::fetchStatus()	{
	return binded == 3 ? _fetch() : FALSE;
	}

Result SyQuery::handleResults()	{
	CS_RETCODE retcode;
	CS_INT res_type;
	while ((retcode = cmd.getResult(&res_type)) == CS_SUCCEED)	{
		if (handleResult(res_type) == CS_FAIL)
			return FAIL;
		if (res_type == CS_ROW_RESULT || res_type == CS_PARAM_RESULT ||
														res_type == CS_STATUS_RESULT)
			return OK;
		}
	return  (retcode == CS_END_RESULTS ? OK : FAIL);
	}

Result SyQuery::_prepare(char* queryText)	{
	if (state & DQS_PREPARED)		//		Already made
		return OK;

	if (!(state & DQS_HAS_STMT))	{
		if (cmd.alloc() == FAIL)
			return FAIL;
		state |= DQS_HAS_STMT;
		}
//	if (isStatic)	{
//		state |= DQS_PREPARED;
//		return OK;
//		}

	MemBuf swap(MemBuf::TEXT_MODE);
	if (flags & SQF_UPPERCASE)	{		//		We need to make object names upper-case
		Bool inQuotes = FALSE;
		for (char* ptr = queryText; *ptr; ptr++)	{
			if (ptr[0] == '|' && ptr[1] == '|')	{
				swap << '+';
				ptr++;
				}
			else	{
				if (*ptr == '\'' || *ptr == '\"')	inQuotes = !inQuotes;
				if (inQuotes)		swap << *ptr;
				else					swap << (char)toupper(*ptr);
				}
			}
		swap << '\0';
		}
	else
		swap << queryText;
	queryText = swap.getBuf();

	//		It's necessary to remove '?' after call of procedure
	Bool removeQuest = FALSE;
	if (flags & SQF_RPC)
		removeQuest = TRUE;
	else	{
		char* ptr, *word1 = strchr(queryText,' ');
		if (word1)
			for (ptr = queryText; ptr < word1; ptr++)
				*ptr = toupper(*ptr);
		ptr = strstr(queryText,"EXECUTE");
		if (ptr && ptr < word1)
			removeQuest = TRUE;
		}
	if (removeQuest)	{
		Bool inQuotes = FALSE;
		for (char* ptr = queryText; *ptr; ptr++)
			if (*ptr == '\'' || *ptr == '\"')	inQuotes = !inQuotes;
			else	if (*ptr == '?' && !inQuotes)	{
				*ptr = 0;
				break;
				}
		for (--ptr; ptr >= queryText; ptr--)
			if (*ptr == ' ')	*ptr = 0;
			else	break;
		}

	if (flags & SQF_STATIC)	{
		if (ct_command(cmd,CS_LANG_CMD,queryText,CS_NULLTERM,CS_UNUSED) != CS_SUCCEED)
			return FAIL;
		}
	else	if (flags & SQF_RPC)	{
		if (ct_command(cmd,CS_RPC_CMD,queryText,CS_NULLTERM,CS_UNUSED) != CS_SUCCEED)
			return FAIL;
		}
	else	{
		if (ct_dynamic(cmd, CS_PREPARE, getCursorName(), CS_NULLTERM,
				queryText,CS_NULLTERM) != CS_SUCCEED || cmd.send() != CS_SUCCEED)
			return FAIL;
		if (handleResults() == FAIL)
			return FAIL;
		}

	state |= DQS_PREPARED;
	return OK;
	}

Result SyQuery::_open(SQLDA* _in_sqlda, short)	{
	if (_execute(_in_sqlda,SYM_OPEN_CURSOR) == FAIL)
		return FAIL;
//	return openCursor();
	return OK;
	}

Result SyQuery::openCursor()	{	return binded ? OK : bind(desc);	}

Result SyQuery::_execute(SQLDA* _in_sqlda, short method)	{
	StatStarter stat;
	if (state & DQS_CACHED)
		stat.init((char*)getQueryName());

	if (flags & SQF_STATIC)	{
		if (cmd.send() != CS_SUCCEED || handleResults() == FAIL)
			return FAIL;
		state |= DQS_PROCESSED;
		}
	else
		if (flags & SQF_RPC)	{
			if (passArg(_in_sqlda) == FAIL || cmd.send() != CS_SUCCEED
														|| handleResults() == FAIL)
				return FAIL;
		state |= DQS_PROCESSED;
		}
	else
		switch (method)	{
			case SYM_OPEN_CURSOR :	{
				if (!(state & DQS_HAS_CURSOR))	{
					if	(ct_dynamic(cmd, CS_CURSOR_DECLARE, getCursorName(),
								CS_NULLTERM, getCursorName(), CS_NULLTERM) != CS_SUCCEED
							||	ct_cursor(cmd, CS_CURSOR_OPTION, 0, CS_UNUSED,
												0, CS_UNUSED, CS_READ_ONLY) != CS_SUCCEED
							|| cmd.send() != CS_SUCCEED
							|| handleResults() == FAIL)
						return FAIL;
					state |= DQS_HAS_CURSOR;
					if (state & DQS_CACHED)	{
						__Query* cached = findByName(getQueryName());
						if (cached)
							cached->state |= DQS_HAS_CURSOR;
						}
					}
				CS_INT res_type;
				if (ct_cursor(cmd, CS_CURSOR_OPEN, 0,
											CS_UNUSED, 0, CS_UNUSED, CS_UNUSED) != CS_SUCCEED
						|| passArg(_in_sqlda) == FAIL
						|| cmd.send() != CS_SUCCEED
						|| cmd.getResult(&res_type) != CS_SUCCEED
						||	handleResult(res_type) != CS_SUCCEED)
					return FAIL;
				state |= DQS_PROCESSED;
				}
				break;

			case SYM_EXECUTE :
				if (ct_dynamic(cmd, CS_EXECUTE, getCursorName(),
											CS_NULLTERM, 0, CS_UNUSED) != CS_SUCCEED
							|| passArg(_in_sqlda) == FAIL
							|| cmd.send() != CS_SUCCEED
							|| handleResults() == FAIL)
					return FAIL;
				state |= DQS_PROCESSED;

			default :;
			}
	return OK;
	}

Bool SyQuery::_fetch()  {
	out_sqlda = out.getSQLDA();
	CS_INT rows_read;
	LstFlush(blobs,SyBlob);
	switch (ct_fetch(cmd,CS_UNUSED,CS_UNUSED,CS_UNUSED,&rows_read))	{
		case CS_END_DATA :
			binded = 0;
			if (handleResults() == FAIL)
				return FAIL;
			return FALSE;
		case CS_SUCCEED :	{
			if (!bindings)		//		Load blobs
				for (CS_INT i = 0; i < out_sqlda->sqld; i++)
					if (isBlob(i))	{
						SyBlob* blob = new SyBlob;
						blob->sqltype = out_sqlda->sqlvar[i].sqltype;
						if (blob->read(this,i+1) == FAIL)	{
							delete blob;
							return FAIL;
							}
						}
			return TRUE;
			}
		}
	return FAIL;
	}

Result SyQuery::_close(Bool deallocate)  {
	if (state & DQS_HAS_CURSOR)	{
		if ((state & DQS_PROCESSED))	{
			if (ct_cursor(cmd, CS_CURSOR_CLOSE, 0, CS_UNUSED, 0, CS_UNUSED,
								deallocate ? CS_DEALLOC : CS_UNUSED) != CS_SUCCEED
					||	cmd.send() != CS_SUCCEED
					|| handleResults() == FAIL)
				return FAIL;
			if (deallocate)
				state &= ~DQS_HAS_CURSOR;
			state &= ~DQS_PROCESSED;
			binded = 0;
			}
		else
			if (deallocate)	{
				if (ct_cursor(cmd, CS_CURSOR_DEALLOC, 0, CS_UNUSED, 0, CS_UNUSED,
									CS_UNUSED) != CS_SUCCEED
							||	cmd.send() != CS_SUCCEED
							|| handleResults() == FAIL)
					return FAIL;
				state &= ~DQS_HAS_CURSOR;
				}
		}
	else
		if (state & DQS_PROCESSED)	{
			if (cmd.trans->client == &cmd)
				if (ct_cancel(NULL,cmd,CS_CANCEL_ALL) != CS_SUCCEED)
//						||	cmd.send() != CS_SUCCEED
//						|| handleResults() == FAIL)
					return FAIL;
			state &= ~DQS_PROCESSED;
			}
	return OK;
	}

Result SyQuery::_unprepare()	{
	binded = 0;
	if ((state & DQS_PREPARED) && !(flags & (SQF_STATIC | SQF_RPC)))
		if (ct_dynamic(cmd,CS_DEALLOC,getCursorName(),
							CS_NULLTERM, 0, CS_UNUSED) != CS_SUCCEED
					||	cmd.send() != CS_SUCCEED
					|| handleResults() == FAIL)
			return FAIL;
	return __Query::_unprepare();
	}

Result SyQuery::close(Bool ignoreError)
  {	return (_close(FALSE) == OK || ignoreError) ? OK : FAIL;	}

Result SyQuery::drop(Bool ignoreError)  {
	if (_close(TRUE) == FAIL)
		return ignoreError ? OK : FAIL;

	if (_unprepare() == FAIL)
		return FAIL;

	if (state & DQS_HAS_STMT)	{
		if (cmd.free() == FAIL)
			return ignoreError ? OK : FAIL;
		state &= ~DQS_HAS_STMT;
		format[0] = Ft_Undefine;
		}
	return OK;
	}

Result SyQuery::release()	{
	if (__Query::release() == FAIL)
		return FAIL;
	cmd.clear();
	binded = 0;
	return OK;
	}

SyBlob* SyQuery::findBlob(CS_INT column)	{
	SyBlob* blob;
	for (LstIterForward(blobs,blob))
		if (blob->column == column)
			return blob;
	return 0;
	}

Result SyQuery::get(char* queryText, ...)	{
	if (queryText)
		getFormat(queryText);
	va_list  ap;
//	MemBuf parBuffer(MemBuf::DATABASE_MODE);

	if (prepare(queryText) == FAIL)
		return FAIL;

	// Accept input parameters
	va_start(ap, queryText);
	fillSQLDA(in,ap);
	va_end(ap);
	in.synchronize();

	if (_open(in,0) == FAIL)
		return FAIL;
	Bool fetched = fetch();
	if (fetched == FALSE)
		return queryError(QUE_NO_ROWS);
	return fetched == FAIL ? FAIL : OK;
	}

Result SyQuery::RPC(int nOut, char* queryText, ...)	{
	flags = SQF_RPC;
	nOutput = nOut;

	if (queryText)
		getFormat(queryText);
	va_list  ap;
	if (prepare(queryText) == FAIL)
		return FAIL;

	// Accept input parameters
	va_start(ap, queryText);
	fillSQLDA(in,ap);
	va_end(ap);
	in.synchronize();

	return _open(in,0);
	}
//	___________________________________________________________________

//	-------------			Parameters treatment					----------
//	________________________________________________________________
Result SyQuery::setArg(MemBuf& dest, short type, SQLVAR* var, va_list& ap)	{
	switch (type)	{
		case Ft_Date_ :	{
			Date* p = va_arg(ap,Date*);
			if (p && DateIsNull(*p))	p = 0;
			if (p == 0)
				var->sqlind = &nullind;
			else	{
				SyDate date;
				date = *p;
				dest << date.dtdays << date.dttime;
				}
			var->sqltype = SQL_DATE;
			break;
			}
		default :
			return __Query::setArg(dest,type,var,ap);
		}
	return OK;
	}

Result SyQuery::setUniArg(DbSQLDA& sqlda, int i, UniVar* uv)	{
	SQLVAR& var = sqlda[i];
	switch (var.sqltype)	{
		case SQL_DATE :	{
			SyDate* date = (SyDate*)var.sqldata;
			date->dtdays = ((UniDT*)uv->value)->days;
			date->dttime = ((UniDT*)uv->value)->secs;
			break;
			}
		case SQL_BLOB :
		case SQL_LTEXT :
		default :
			return __Query::setUniArg(sqlda,i,uv);
		}
	return OK;
	}

Result SyQuery::getArg(DbSQLDA& sqlda, void* data, short type, int i)	{
	out_sqlda = sqlda.getSQLDA();
	switch (type)	{
		case Ft_Date_ :
			if (isFieldNull(i))			*((Date*)data) = Date::errorDate;
			else			*((Date*)data)=(Date)*(SyDate*)getFieldValue(i);
			break;
		case Ft_Quad_ :
			if  (isFieldNull(i))
				((SyBlobId*)data)->isNotNull = 0;
			else	{
				((SyBlobId*)data)->isNotNull = 1;
				((SyBlobId*)data)->blob = *findBlob(1+i);
				}
			((SyBlobId*)data)->isNotModified = 1;
			break;

		case Ft_Binary_ :	{		//		Redefine
			if (getFieldType(i) != SQL_BLOB)
				return __Query::getArg(sqlda,data,type,i);

			Dump* dump = (Dump*)data;
			if (isFieldNull(i))
				dump->destroy();
			else	{
				char* p = (char*)getFieldValue(i);
				textToBinary(*dump,p);
				}
			break;
			}

		default :
			return __Query::getArg(sqlda,data,type,i);
		}
	return OK;
	}

Result SyQuery::getUniArg(DbSQLDA& sqlda, int i, UniVar* uv)	{
	SQLVAR& var = sqlda[i];
	uv->uniType = var.sqltype;
	switch (var.sqltype)	{
		case SQL_DATE :
			uv->value = var.sqldata;
			break;
		default :
			return __Query::getUniArg(sqlda,i,uv);
		}
	return OK;
	}

char* SyQuery::getTextField(int fldNo)	{
	char* ptr = (char*)getFieldValue(fldNo);
	if (isFieldNull(fldNo))	*ptr = 0;
	else
		for (char* tail = ptr + strlen(ptr) - 1; tail >= ptr && *tail == ' '; tail--)
			*tail = '\0';
	return ptr;
	}

Bool SyQuery::isBlob(int fldNo)	{
	short sqltype = out_sqlda->sqlvar[fldNo].sqltype;
	return sqltype == SQL_LTEXT || sqltype == SQL_BLOB;
	}

char SyQuery::semiByteToChar(char b)	{
	return (b < 0xA ? '0' + b : 'a' + (b - 0xA));
	}

char SyQuery::charToSemiByte(char c)	{
	c = tolower(c);
	return (c >= 'a' ? 0xA + (c - 'a') : c - '0');
	}

void SyQuery::binaryToText(MemBuf& dest, int size, char* data)	{
	if (size == 0)		dest << "NULL";
	else					dest << "0x";
	for (int i = 0; i < size; i++)
		dest << StdFormat("%c%c", semiByteToChar((data[i] >> 4)&0xF),
									semiByteToChar(data[i]&0xF));
	}

void SyQuery::textToBinary(Dump& dest, char* text)	{
	int size = strlen(text)/2;
	dest.create(size);
	for (int j = 0; text[j*2]; j++)
		dest.data[j] = (charToSemiByte(text[j*2]) << 4)|charToSemiByte(text[j*2+1]);
	}
//	________________________________________________________________

//	-------------			Sybase query extension				----------
//	________________________________________________________________
SQLDA* SyQuery::describeInput()	{
	if (ct_dynamic(cmd, CS_DESCRIBE_INPUT, getCursorName(), CS_NULLTERM,
					0,CS_UNUSED) != CS_SUCCEED || cmd.send() != CS_SUCCEED)
		return 0;
	if (handleResults() == FAIL)
		return 0;
	return desc;
	}

Result SyQuery::staticExecute(char *text)	{
	if (cmd.alloc() == FAIL)
		return FAIL;

	if (ct_command(cmd,CS_LANG_CMD,text,CS_NULLTERM,CS_UNUSED) != CS_SUCCEED
						|| cmd.send() != CS_SUCCEED	|| handleResults() == FAIL)
		return FAIL;
	return OK;
	}

Result SyQuery::staticExecute1(char *text, SQLDA* _in_sqlda)	{
	if (cmd.alloc() == FAIL)
		return FAIL;

	if (ct_command(cmd,CS_LANG_CMD,text,CS_NULLTERM,CS_UNUSED) != CS_SUCCEED
						|| passArg(_in_sqlda) == FAIL
						|| cmd.send() != CS_SUCCEED	|| handleResults() == FAIL)
		return FAIL;
	return OK;
	}

Result SyQuery::staticExecute2(char *text, char* data)	{
	SQLDA s;
	SQLDA* sqlda = &s; //(SQLDA*)StdMalloc(SQLDA_LENGTH(1));
	sqlda->sqld = sqlda->sqln = 1;
	sqlda->sqlvar[0].sqltype = CS_TEXT_TYPE;
	if (data)
		sqlda->sqlvar[0].sqldata = data;
	else
		sqlda->sqlvar[0].sqlind = &nullind;
	return staticExecute1(text,sqlda);
	}
//	________________________________________________________________

//	-------------			SyBlob implementation				----------
//	________________________________________________________________
Result SyBlob::write()	{
	if (column == -1)
		return queryError(QUE_BAD_BLOBID);
	SyQuery q;
	if (q.cmd.alloc() == FAIL)
		return FAIL;

	Result result = FAIL;
	desc.total_txtlen = size;
	desc.log_on_update = 1;
	char nothing;
	if (size == 0)
		data = &nothing;
	else if (sqltype == SQL_LTEXT)
		CpgConvertString(data,Cpg_CP_DEFAULT,data,Cpg_CP_DATABASE,size);
	CS_INT res_type;
	if (ct_command(q.cmd,CS_SEND_DATA_CMD,0,CS_UNUSED,CS_COLUMN_DATA) == CS_SUCCEED)
		if (ct_data_info(q.cmd,CS_SET,CS_UNUSED,&desc) == CS_SUCCEED)
			if (ct_send_data(q.cmd,data,size) == CS_SUCCEED)
				if (q.cmd.send() == CS_SUCCEED)
					if (q.cmd.getResult(&res_type) == CS_SUCCEED)
						if (q.handleResult(res_type) == CS_SUCCEED)
							result = OK;
	if (size == 0)
		data = 0;
	else if (sqltype == SQL_LTEXT)
		CpgConvertString(data,Cpg_CP_DATABASE,data,Cpg_CP_DEFAULT,size);
	if (result == OK)	{
//		if (q.openCursor() == FAIL)		//		To get new descriptor
//			return FAIL;
		while ((result = q.fetch()) == TRUE)	//		It must be only one
			memcpy(desc.timestamp,q.getFieldValue(0),CS_TS_SIZE);
		if (result != FAIL)
			result = OK;
		}
	if (q.cmd.free() == FAIL)
		result = FAIL;
	return result;
	}

Result SyBlob::read(SyQuery* owner, CS_INT index, Bool dontLoad)	{
	CS_RETCODE retcode = ct_get_data(owner->cmd,index,&column,0,0);
	if (retcode != CS_SUCCEED && retcode != CS_END_ITEM
			&& retcode != CS_END_DATA)
		return FAIL;

/*
	int textLimit;
	if (ct_con_props(owner->cmd.trans->connection, CS_GET, CS_TEXTLIMIT,&textLimit,
						CS_UNUSED, NULL)  != CS_SUCCEED)
		return FAIL;
*/

	if (ct_data_info(owner->cmd,CS_GET,index,&desc) != CS_SUCCEED)
		return FAIL;
	if (!dontLoad)	{
		create(desc.total_txtlen,0);
		if (size > 0)	{
			retcode = ct_get_data(owner->cmd,index,data,size,0);
			if (retcode != CS_END_ITEM	&& retcode != CS_END_DATA)
				return FAIL;
			if (sqltype == SQL_LTEXT)
				CpgConvertString(data,Cpg_CP_DATABASE,data,Cpg_CP_DEFAULT,size);
			}
		}
	if (!owner->blobs.isContains(this))
		owner->blobs.Add(this);
	column = index;
	return OK;
	}

SyBlob& SyBlob::operator=(const SyBlob& src)	{
	Dump::operator=((const Dump&)src);
	desc = src.desc;
	column = src.column;
	sqltype = src.sqltype;
	return *this;
	}

SyBlobId& SyBlobId::operator=(const SyBlobId& id)	{
	isNotNull = id.isNotNull;
	isNotModified = 0;
	if (id.blob.column != -1)
		blob = id.blob;
	return *this;
	}

int SyBlobId::operator==(const SyBlobId& id)
	{	return isNotNull == 0 && id.isNotNull == 0;	}

void SyBlobId::get(MemBuf& buf)	{
	if (isNotNull && blob.size)
		buf.fill(blob.data,blob.size);
	}

void SyBlobId::get(Dump& buf)	{
	if (isNotNull && blob.size)
		buf = blob;
	else
		buf.destroy();
	}

void SyBlobId::set(const MemBuf& buf)	{
	isNotNull = 1;
	isNotModified = 0;
	blob.create(buf.getLength(),buf.getBuf());
	}

void SyBlobId::set(const Dump& buf)	{
	isNotNull = 1;
	isNotModified = 0;
	blob.Dump::operator=(buf);
	}

Result SyBlobId::update()	{
	if (isNotModified)
		return OK;
	if (!isNotNull)
		blob.destroy();
	return blob.write();
	}
//	________________________________________________________________

Result SyBlob::play(Dump& data, char* queryText, ...)	{
	SyQuery q;
	q.getFormat(queryText);

	va_list  ap;
	va_start(ap, queryText);
	q.in.clear();
	q.fillSQLDA(q.in,ap);
	va_end(ap);
	q.in.synchronize();
	ExactString req(queryText);
	data << (Dump&)q.in.sqlda << (Dump&)q.in.record << (Dump&)req << (Dump&)*this;
	return write();
	}

Result SyBlob::replay(Dump& data)	{
	SyQuery q;
	ExactString req;
	Dump swap;
	data >> swap >> (Dump&)req >> (Dump&)q.in.record >> (Dump&)q.in.sqlda;

	if (q.prepare(req) == FAIL)
		return FAIL;

	q.in.synchronize();
	if (q._open(q.in) == FAIL)
		return FAIL;

	Bool fetched;
	while ((fetched = q.fetch()) == TRUE)	{
		SyBlob a000;
		if (a000.read(&q,0,TRUE) == FAIL)
			return FAIL;
		a000.Dump::operator=(swap);
		if (a000.write() == FAIL)
			return FAIL;
		}
	return (fetched == FAIL ? FAIL : OK);
	}

