#include "__queryf.hpp"
#include "__query.hpp"
#include "membuf.h"
#include "list.h"
#include "timer.h"
#include "dump.hpp"
#include <stdio.h>

char* queryErrorMessages[] = {
	"%s",
	"Field %s is not found in query %s",
	"Can not unprepare cached statement %s",
	"This transaction can not be activated",
	"Column %d isn't text or image",
	"Blob id is invalid",
	"Cannot disconnect while transaction is still active",
	"Cannot activate connection/transaction",
	"Cannot start transaction - connection is closed",
	"Cannot start transaction - already started",
	"Cannot commit transaction - not started",
	"Cannot rollback transaction - not started",
	"No rows returned",
	"Redefinition of query %s :\n%s\nwith %s",
	"Query name %s is too long (Max = 10)",
	"Value is to be truncated : %s",
	"Type is not supported : %d",
	0	};

ErrorClass queryError("queryError",queryErrorMessages);

short __Query::nullind = -1;

//			-----------			SQLDA utilities			--------------------
//			______________________________________________________________

//		Fills element with SQLVAR <var> (assuming that there's enough room after
//		<text> to store <var> value
DbRecordField& DbRecordField::operator=(SQLVAR* var)	{
	length = sizeof(DbRecordField) + var->sqllen;
	if (var->sqlind)
		nullind = *var->sqlind;
	else
		nullind = 0;
	if (var->sqldata && nullind != __Query::nullind)
		memcpy(text,var->sqldata,var->sqllen);
	text[var->sqllen] = 0;
	return *this;
	}

DbRecordField* DbRecord::extend(int delta)	{
	int last = used;
	used += delta;
	if (used <= size)	{}
	else	if (size > 0)	{
		Dump swap(*this);
		Dump::create(used + FIELD_GRAN*AVG_FIELD_SIZE);
		memcpy(data,swap.data,swap.size);
		}
	else
		Dump::create(used + FIELD_GRAN*AVG_FIELD_SIZE);
	return (DbRecordField*)(data+last);
	}

void DbRecord::create(SQLDA* sqlda)	{
	nField = sqlda->sqld;
	int size = sizeof(DbRecordField)*sqlda->sqld;
	for (int i = 0; i < sqlda->sqld; i++)
		size += sqlda->sqlvar[i].sqllen;
	extend(size);
	int offset = 0;
	for (i = 0; i < sqlda->sqld; i++)	{
		SQLVAR* var = &sqlda->sqlvar[i];
		DbRecordField* field = (DbRecordField*)(data+offset);
		*field = var;
		offset += field->length;
		}
	}

//		Gets record buffer element for field <i>
DbRecordField* DbRecord::get(int i)	{
	int offset = 0, count = 0;;
	while (count < nField)	{
		DbRecordField* field = (DbRecordField*)(data+offset);
		if (count == i)
			return field;
		offset += field->length;
		count++;
		}
	return 0;
	}

DbRecord& DbRecord::operator<<(SQLVAR* var)	{
	int need = sizeof(DbRecordField) + var->sqllen;
	DbRecordField* field = extend(need);
	nField++;
	*field = var;
	return *this;
	}

void DbRecord::clear()	{
	destroy();
	used = 0;
	nField = 0;
	}

void DbSQLDA::extend(int delta)	{
	int used;
	if (sqlda.size > 0)
		used = SQLDA_LENGTH(getSQLDA()->sqld) + delta*sizeof(SQLVAR);
	else
		used = SQLDA_LENGTH(delta);
	if (used > sqlda.size)
		if (sqlda.size > 0)	{
			Dump swap(sqlda);
			sqlda.create(used + FIELD_GRAN*sizeof(SQLVAR));
			memcpy(sqlda.data,swap.data,swap.size);
			getSQLDA()->sqln += delta;
			}
		else	{
			sqlda.create(used + FIELD_GRAN*sizeof(SQLVAR));
			getSQLDA()->sqln = delta;
			}
	else
		getSQLDA()->sqln += delta;
	}

DbSQLDA& DbSQLDA::operator<<(SQLVAR* var)	{
	extend(1);
	if (hasRecord)
		record << var;
	(*this)[getSQLDA()->sqld++] = *var;
	return *this;
	}

void DbSQLDA::synchronize()	{
	if (!hasRecord)
		return;

	SQLDA* s = getSQLDA();
	for (int i = 0; i < s->sqld; i++)	{
		DbRecordField* field = record.get(i);
		s->sqlvar[i].sqldata = field->text;
		s->sqlvar[i].sqlind = &field->nullind;
		}
	}

void DbSQLDA::create(SQLDA* s, Bool withData)	{
	sqlda.create(SQLDA_LENGTH(s->sqld));
	memcpy(sqlda.data,s,sqlda.size);
	getSQLDA()->sqln = s->sqld;
	if (withData)	{
		hasRecord = TRUE;
		record.create(s);
		synchronize();
		}
	else
		hasRecord = FALSE;
	}

void DbSQLDA::clear()	{
	sqlda.destroy();
	record.clear();
	hasRecord = TRUE;
	}
//			______________________________________________________________

//			-----------		Format parsing system		--------------------
//			______________________________________________________________
short __Query::typeToShort(char* ptr, int& size)	{
	short ts = 0;	//		no match
	size = 0;
	if (strchr("BSsFfDQEeIiURM",ptr[0]))	{
		ts = CC2(ptr[0],' ');
		size = 1;
		}
	else
		if (ptr[0] == 'h' && strchr("DdEe",ptr[1]) ||
				ptr[0] == 'l' && strchr("DdEeFf",ptr[1]))	{
			ts = CC2(ptr[0],ptr[1]);
			size = 2;
			}
	return ts;
	}

void getFormat(short* format, MemBuf& clearQuery)	{
	char* start = clearQuery.getBuf();
	int lastParam = 0;
	for (char* ptr = strchr(start,'%'); ptr; ptr = strchr(ptr,'%'))	{
		if (ptr[1] == '%')	{
			clearQuery.squeze(1,(int)(ptr-start));
			ptr++;
			}
		else	{
			int toSqueeze = 0;
			short ts = __Query::typeToShort(ptr+1,toSqueeze);
			if (toSqueeze)	{
				*ptr = '?';
				clearQuery.squeze(toSqueeze,(int)(ptr-start+1));
				format[lastParam++] = ts;
				}
			else
				ptr++;
			}
		}
	format[lastParam] = Ft_Stop;
	}

void __Query::getFormat(char* queryText)	{
	if (format[0] == Ft_Undefine)	{
		MemBuf clearQuery;
		clearQuery << queryText;
		::getFormat(format,clearQuery);
		}
	}

char* __Query::preRead(short& fieldNo, short& type, char* ptr)	{
	fieldNo++;			//		Next field no by default
	type = 0;
	//		Try to extract type
	if (ptr[1] != '#' && ptr[1] != '(' && ptr[1] != '%')	{
		int toSqueeze = 0;
		type = typeToShort(ptr+1,toSqueeze);
		ptr += toSqueeze;
		}
	switch (ptr[1])	{
		case '#' :	//		pointed by ordinal number
			sscanf(ptr+2,"%d",&fieldNo);
			break;
		case '(' :	{	//		pointed by field name
			char* tail = strchr(ptr+2,')');
			char name[MAXSTR];
			strncpy(name,ptr+2,(int)(tail-ptr-2));
			name[(int)(tail-ptr-2)] = 0;
			fieldNo = getFieldNo(name);
			break;
			}
		}
	ptr++;

	//		Now we have fieldNo
	if (type == 0)
		switch (getFieldType(fieldNo))	{
			case SQL_SHORT	:	type = Ft_short_;	break;
			case SQL_LONG 	:	type = Ft_long_;		break;
			case SQL_FLOAT	:	type = Ft_float_;	break;
			case SQL_DOUBLE:	type = Ft_double_;	break;
			case SQL_DATE 	:	type = Ft_Date_;		break;
			default :			type = Ft_FString_;	break;
			}
	return ptr;
	}
//			______________________________________________________________


//			--------		Common SQLDA processing mechanism		-----------
//			______________________________________________________________

//		Following class implements acceleror of the field-by-name access
class KeyedFieldName : public DynArrayItem	{
public:
	KeyedFieldName( int fldNo ) { fieldNo = fldNo; }
	int fieldNo;
	};

char* __Query::getFieldName(int fldNo)
	{	return  out_sqlda->sqlvar[fldNo].sqlname;  }

int __Query::getFieldNo(const char *fldName)  {
	int fldNo;
	KeyedFieldName *key;

	if( !fieldByNameAccelerator )	{
		fieldByNameAccelerator = new DynArray;
		fieldByNameAccelerator->OwnsElements(TRUE);
		for( fldNo = 0; fldNo < out_sqlda->sqld; fldNo++ )	 {
			key = new KeyedFieldName(fldNo);
			strupr(out_sqlda->sqlvar[fldNo].sqlname);
			fieldByNameAccelerator->add( out_sqlda->sqlvar[fldNo].sqlname,*key);
			}
		}
	char buf[MAXSTR];
	strcpy(buf, fldName);
	strupr(buf);
	if ((key = (KeyedFieldName*)&fieldByNameAccelerator->find(buf)) != NULL)
		return key->fieldNo;

	return queryError(QUE_FIELD_NOT_FOUND,fldName,getQueryName());
	}

int __Query::getFieldType(int fldNo)
	{	return  out_sqlda->sqlvar[fldNo].sqltype;  }

int __Query::getFieldType(const char *fldName)  {
	int fldNo;
	if( (fldNo = getFieldNo(fldName)) == FAIL)
		return  FAIL;
	return  getFieldType(fldNo);
	}

int __Query::getFieldLength(int fldNo)
	{	return  out_sqlda->sqlvar[fldNo].sqllen;  }

int __Query::getFieldLength(const char *fldName)  {
	int fldNo;
	if ((fldNo = getFieldNo( fldName)) == FAIL)
		return  FAIL;
	return  getFieldLength(fldNo);
	}

Bool __Query::isFieldNull(int fldNo)
	{	return  out_sqlda->sqlvar[fldNo].sqlind &&
						*out_sqlda->sqlvar[fldNo].sqlind;  }

Bool __Query::isFieldNull(const char *fldName)  {
	int fldNo;
	if ((fldNo = getFieldNo( fldName)) == FAIL)
		return FAIL;
	return  isFieldNull(fldNo);
	}

Pointer __Query::getFieldValue(int fldNo)
	{	return  out_sqlda->sqlvar[fldNo].sqldata;	}

Pointer __Query::getFieldValue(const char *fldName)	{
	int   fldNo;

	if ((fldNo = getFieldNo(fldName)) == FAIL)
		return  (Pointer)FAIL;
	return  getFieldValue( fldNo );
	}

char* __Query::getTextField(int fldNo)	{	return (char*)getFieldValue(fldNo);	}
//			______________________________________________________________


//	-----------      	Cacheable query implementation		----------
//	________________________________________________________________

//	------------      Query stack class		-------------------------
//	-------		 (only for internal purposes)			----------------
#define CHECK_QUERY_LENGTH	128
class StoredQuery : public LstLink0	{
public :
	__Query* data;
	ExactString stackName;
	ExactString queryText;
	StoredQuery(const __Query& query, char* text) : LstLink0()	{
		stackName = query.getStackName();
		if (strlen(text) > CHECK_QUERY_LENGTH)
			queryText = StdFormat("%.*s",CHECK_QUERY_LENGTH,text);
		else
			queryText = text;
		data = query.dup();
		data->cloneFrom(query);
		data->state = DQS_PREPARED | DQS_CACHED | DQS_HAS_STMT;
		}
	~StoredQuery()	{
		data->state &= ~DQS_CACHED;
		delete data;
		}
	};

class StackOfQuery : public LstList	{
public :
	StackOfQuery() : LstList("Query Stack")	{};
	~StackOfQuery()	{	flush();	}
	Result Add(const __Query& query, char* text);
															//		Add the query to be
															//		stored. It will be
															//		copied, while adding,
															//		so you have not bother
															//		to keep any query data
															//		NOTE : all queries are
															//		stored prepared but
															//		non-opened
	Result Remove(const char* name);				//		Cancel storing the query
															//		with given name
	Result RemoveStack(const char* name);	 	//		Cancel storing all the query
															//		of the named stack
	StoredQuery* findByName(const char* name);
	static void flush();
	};
StackOfQuery stackOfQuery;			//		Support of stack of stored queries

void StackOfQuery::flush()	{
	for (Pointer item = stackOfQuery.getFirst(); item; item = stackOfQuery.getFirst())	{
		stackOfQuery.LstList::Detach(item);
		delete (StoredQuery*)item;
		}
	}

StoredQuery* StackOfQuery::findByName(const char* name)	{
	for (Pointer item = getFirst(); item; item = getNext(item))
		if (strcmp(((StoredQuery*)item)->data->getQueryName(),name) == 0)
			return (StoredQuery*)item;
	return 0;
	}

__Query* __Query::findByName(const char* name)	{
	StoredQuery* item = stackOfQuery.findByName(name);
	return item ? item->data : 0;
	}


Result StackOfQuery::Add(const __Query& query, char* text)	{
	if (!(query.state & DQS_PREPARED) || findByName(query.getQueryName()))
		return FAIL;
	StoredQuery* storedQuery = new StoredQuery(query,text);
	return LstList::Add(storedQuery);
	}

Result StackOfQuery::Remove(const char* name)	{
	StoredQuery* storedQuery = findByName(name);
	if (!storedQuery)
		return FAIL;
	if (LstList::Detach(storedQuery) == FAIL)
		return FAIL;
	delete storedQuery;
	return OK;
	}

Result StackOfQuery::RemoveStack(const char* name)	{
	for (Pointer next, item = getFirst(); item; item = next)	{
		next = getNext(item);
		StoredQuery* q = (StoredQuery*)item;
		if (strcmp(q->stackName,name) == 0)	{
			LstList::Detach(q);
			delete q;
			}
		}
	return OK;
	}

//		-----		Implementation of __Query class members	----------------
//		---------		providing query storage system		----------------
Result __Query::store(char* text)
	{	return stackOfQuery.Add(*this,text);	}

Bool __Query::restore(char* text)	{
	StoredQuery* storedQuery = stackOfQuery.findByName(getQueryName());
	if (!storedQuery)
		return FALSE;
	if (text)	{
		int ln = min(storedQuery->queryText.size - 1,strlen(text));
		char* old = ln > 0 ? storedQuery->queryText.data : "";
		if (ln <= 0 || memcmp(old,text,ln) != 0)
			return queryError(QUE_REDEFINE,getQueryName(),old,text);
		}
	cloneFrom(*storedQuery->data);
	return TRUE;
	}

Result __Query::dropQuery(const char* name)
	{	return stackOfQuery.Remove(name);	}

Result __Query::dropStack(const char* name)
	{	return stackOfQuery.RemoveStack(name);	}
//	________________________________________________________________


//	-------------      	Common __Query methods				----------
//	________________________________________________________________
__Query::__Query(const char *name)	{
	state = 0;
	if (name && *name)
		setName(name);
	else
		sprintf(queryName,"_%p",this);
	out_sqlda = NULL;
	format[0] = Ft_Undefine;
	fieldByNameAccelerator = 0;	//		Accelerator will be created
											//		during first <getFieldNo> call
	}

//		Copy the state of query that <query> have had just after
//		executing of prepare into this
void __Query::cloneFrom(const __Query& query)	{
	state = DQS_PREPARED | DQS_CACHED | DQS_HAS_STMT;
	memcpy(queryName,query.queryName,sizeof(queryName));
	fieldByNameAccelerator = query.fieldByNameAccelerator;
	memcpy(format,query.format,sizeof(format));
	}

void __Query::setName(const char *name)	{
	state = DQS_CACHED;
	int ln = strlen(name);
	memcpy(queryName,name,min(ln+1,sizeof(queryName)));
	queryName[sizeof(queryName)-1] = 0;
	if (Statistics::find(getQueryName()) == 0)
		new Statistics(getQueryName());
	}
//	________________________________________________________________


//	-------------			Query handling methods				----------
//	________________________________________________________________
Result __Query::prepare(char* queryText, Bool useAsIs)	{
	Bool needToBeStored = FALSE;
	if (state & DQS_PREPARED)
		return OK;

	MemBuf clearQuery;
	if (!useAsIs && queryText)	{
		clearQuery << queryText;
		::getFormat(format,clearQuery);
		queryText = clearQuery.getBuf();
		}
	else
		format[0] = Ft_Stop;

	if (state & DQS_CACHED)					//		First we check if the same query
		switch (restore(queryText))	{	//		has been already stored
			case FAIL :	return FAIL;		//		Store the new statement
			case TRUE :	return OK;
			default :	needToBeStored = TRUE;
			}

	if (_prepare(queryText) == FAIL)
		return FAIL;
	if (needToBeStored)
		if (store(queryText) == FAIL)
			return FAIL;
	return OK;
	}

Result __Query::_unprepare()	{
	state &= ~DQS_PREPARED;
	format[0] = Ft_Undefine;
	return OK;
	}

Result __Query::unprepare()	{
	if (state & DQS_CACHED)
		return queryError(QUE_CANT_UNPREPARE,getQueryName());

	if (state & DQS_OPENED)
		if (close() == FAIL)
			return FAIL;

	return _unprepare();
	}

Result __Query::release()  {
	if (state & DQS_CACHED)
		close(TRUE);
	else
		drop(TRUE);
	StoredQuery* query = stackOfQuery.findByName(getQueryName());
	if (!query)
		delete fieldByNameAccelerator;
	else
		if (fieldByNameAccelerator && !query->data->fieldByNameAccelerator)
			query->data->fieldByNameAccelerator = fieldByNameAccelerator;
	fieldByNameAccelerator = 0;
	format[0] = Ft_Undefine;
	state &= DQS_CACHED;
	return OK;
	}

Result __Query::open(char* queryText, ...)	{
	if (queryText)
		getFormat(queryText);

	va_list  ap;
	va_start(ap, queryText);
	in.clear();
	fillSQLDA(in,ap);
	va_end(ap);
	in.synchronize();

	if (close() == FAIL)						return FAIL;
	if (prepare(queryText) == FAIL)		return FAIL;
	return _open(in,0);
	}

Result __Query::execute(char* queryText, ...)	{
	if (prepare(queryText) == FAIL)
		return FAIL;

	va_list  ap;
	va_start(ap, queryText);
	in.clear();
	fillSQLDA(in,ap);
	va_end(ap);
	in.synchronize();
	return _execute(in);
	}

Bool __Query::fetch(char* format, ...)	{
	Bool result = _fetch();
	if (result == TRUE)	{
		va_list  ap;
		va_start(ap, format);
		if (_read(format,ap) == FAIL)
			result = FAIL;
		va_end(ap);
		}
	return result;
	}

Bool __Query::fetch()	{	return _fetch();	}

Result __Query::read(char* format, ...)	{
	va_list  ap;
	va_start(ap, format);
	Result result = _read(format,ap);
	va_end( ap );
	return result;
	}
//	________________________________________________________________

//	-------------			Parameters treatment					----------
//	________________________________________________________________
void __Query::fillSQLDA(DbSQLDA& sqlda, va_list& ap)	{
	// Calculate number of input parameters
	for (int argc = 0; format[argc] != Ft_Stop; argc++);
	sqlda.extend(argc);				//		It only extends SQLDA struture
	SQLVAR var;

	MemBuf dest(MemBuf::DATABASE_MODE);
	for (argc = 0; format[argc] != Ft_Stop; argc++)	{
		dest.flush();
		var.sqlind = 0;
		setArg(dest, format[argc], &var, ap);
		var.sqllen = (var.sqlind == &nullind) ? 0 : dest.getLength();
		if (var.sqllen > 0 && (var.sqltype & ~1) == SQL_FIXCHAR
															&& format[argc] != Ft_Binary_)
			var.sqllen--;
		var.sqldata = dest.getBuf();
		sqlda << &var;
		}
	}

Result __Query::setArg(MemBuf& dest, short type, SQLVAR* var, va_list& ap)	{
	switch (type)	{
		case Ft_FString_ :
		case Ft_char_ :
		case Ft_hEnum : case Ft_Enum : case Ft_lEnum :
		case Ft_hEnum_ : case Ft_Enum_ : case Ft_lEnum_ :
		case Ft_dump_ : case Ft_quark_ :
			switch(type)	{
				case Ft_FString_ :	case Ft_char_ :
				case Ft_dump_ : case Ft_quark_ :
				case Ft_membuf_ :	{
					char* p = va_arg(ap,char*);
					if (!p)	var->sqlind = &nullind;
					else
						switch (type)	{
							case Ft_FString_ : dest << ((FString8*)p)->get();	break;
							case Ft_dump_ :	dest << (char*)(((Dump*)p)->data);	break;
							case Ft_quark_ :	dest << (char*)*(Quark*)p;		break;
							case Ft_membuf_ :	{
								MemBuf* mp = (MemBuf*)p;
								if (mp->getLength())
									dest << mp->getBuf();
								else
									dest << "";
								break;
								}
							default :			dest << p;
							}
					break;
					}
				case Ft_hEnum :	storeEnum(dest,(short)va_arg(ap,int));	break;
				case Ft_Enum :		storeEnum(dest,(short)va_arg(ap,int));	break;
				case Ft_lEnum :	storeEnum(dest,(short)va_arg(ap,long));break;
				case Ft_hEnum_ :	case Ft_Enum_ :	case Ft_lEnum_ :	{
					void* p = va_arg(ap,void*);
					if (!p)	var->sqlind = &nullind;
					else	switch (type)	{
						case Ft_hEnum_ :	storeEnum(dest,*(short*)p);		break;
						case Ft_Enum_ :	storeEnum(dest,(short)*(int*)p); break;
						case Ft_lEnum_ :	storeEnum(dest,(short)*(long*)p); break;
						}
					}
				}
			var->sqltype = SQL_FIXCHAR;
			break;

		case Ft_short : case Ft_bool :	//		Short is always passing as int
			dest << (short)va_arg(ap,int);
			var->sqltype = SQL_SHORT;
			break;
		case Ft_short_ :	{
			short* p = va_arg(ap,short*);
			if (p)		dest << *p;
			else				var->sqlind = &nullind;
			var->sqltype = SQL_SHORT;
			break;
			}
		case Ft_bool_ :	{
			int* p = va_arg(ap,int*);
			if (p)		dest << (short)*p;
			else				var->sqlind = &nullind;
			var->sqltype = SQL_SHORT;
			break;
			}
		case Ft_long :
			dest << va_arg(ap,long);
			var->sqltype = SQL_LONG;
			break;
		case Ft_long_ :	{
			long* p = va_arg(ap,long*);
			if (p)		dest << *p;
			else				var->sqlind = &nullind;
			var->sqltype = SQL_LONG;
			break;
			}
		case Ft_float :		// 	Float is always passing as double
			dest << (float)va_arg(ap,double);
			var->sqltype = SQL_FLOAT;
			break;
		case Ft_float_ :	{
			float* p = va_arg(ap,float*);
			if (p)		dest << *p;
			else	var->sqlind = &nullind;
			var->sqltype = SQL_FLOAT;
			break;
			}
		case Ft_double :
			dest << va_arg(ap,double);
			var->sqltype = SQL_DOUBLE;
			break;
		case Ft_double_ :	{
			double* p = va_arg(ap,double*);
			if (p)		dest << *p;
			else				var->sqlind = &nullind;
			var->sqltype = SQL_DOUBLE;
			break;
			}
		case Ft_Binary_ :	{
			Dump* p = va_arg(ap,Dump*);
			if (!p)	var->sqlind = &nullind;
			else
				for (int i = 0; i < p->size; i++)
					dest << p->data[i];
			var->sqltype = SQL_VARCHAR;
			break;
			}
		}
	return OK;
	}

Result __Query::_read(char* format, va_list ap)	{
	short fieldNo = -1;
	for( char* ptr = strchr(format,'%'); ptr; ptr = strchr(ptr,'%'))	{
		short type = 0;
		ptr = preRead(fieldNo,type,ptr);

		void* data = va_arg(ap,void*);

		if (getArg(out,data,type,fieldNo) == FAIL)
			return FAIL;
		}
	return OK;
	}

Result __Query::getArg(DbSQLDA& sqlda, void* data, short type, int i)	{
	out_sqlda = sqlda.getSQLDA();
	switch (type)	{
		case Ft_FString_ :
			if (isFieldNull(i))		*(FString8*)data = "";
			else	CpgConvertString(getTextField(i),Cpg_CP_DATABASE,
											(char*)*(FString8*)data,Cpg_CP_DEFAULT,0);
			break;
		case Ft_char_ :
			if (isFieldNull(i))	*(char*)data = 0;
			else	CpgConvertString(getTextField(i),Cpg_CP_DATABASE,
											(char*)data,Cpg_CP_DEFAULT,0);
			break;
		case Ft_dump_ :
		case Ft_quark_ :
		case Ft_membuf_ :	{
			Dump dump;
			if (isFieldNull(i))
				dump.create(1,"");
			else	{
				char* value = (char*)getTextField(i);
				dump.create(strlen(value)+1,value);
				CpgConvertString((char*)dump.data,Cpg_CP_DATABASE,
										(char*)dump.data,Cpg_CP_DEFAULT,0);
				}
			if (type == Ft_dump_)					*(Dump*)data = dump;
			else if (type == Ft_quark_)	*(Quark*)data = (const char*)dump.data;
			else	{
				MemBuf* mp = (MemBuf*)data;
				mp->flush();
				if (dump.size > 1)
					mp->insert(dump.data,dump.size);
				else if (mp->dataMode & MemBuf::TEXT_MODE)
					*mp << '\0';
				}
			break;
			}
		case Ft_hEnum : case Ft_Enum : case Ft_lEnum :
		case Ft_hEnum_ : case Ft_Enum_ : case Ft_lEnum_ :	{
			char buffer[3];
			CpgConvertString(isFieldNull(i) ? "" : getTextField(i),
										Cpg_CP_DATABASE, buffer,Cpg_CP_DEFAULT,2);
			buffer[2] = 0;
			switch (type)	{
				case Ft_hEnum : case Ft_hEnum_ :
					*(short*)data = (short)DbStoE(buffer);
					break;
				case Ft_Enum : case Ft_Enum_ :
					*(int*)data = (int)DbStoE(buffer);
					break;
				case Ft_lEnum : case Ft_lEnum_ :
					*(long*)data = (long)DbStoE(buffer);
					break;
				}
			break;
			}
		case Ft_short : case Ft_short_ :
			if (isFieldNull(i))	*(short*)data = 0;
			else			*(short*)data = *(short*)getFieldValue(i);
			break;
		case Ft_bool : case Ft_bool_ :
			if (isFieldNull(i))	*(int*)data = 0;
			else			*(int*)data = *(short*)getFieldValue(i);
			break;
		case Ft_long : case Ft_long_ :
			if (isFieldNull(i))	*(long*)data = 0;
			else			*(long*)data = *(long*)getFieldValue(i);
			break;
		case Ft_float : case Ft_float_ :
			if (isFieldNull(i))	*(float*)data = 0;
			else			*(float*)data = *(float*)getFieldValue(i);
			break;
		case Ft_double : case Ft_double_ :
			if (isFieldNull(i))	*(double*)data = 0;
			else			*(double*)data = *(double*)getFieldValue(i);
			break;
		case Ft_Binary_ :	{
			Dump* dump = (Dump*)data;
			if (isFieldNull(i))
				dump->destroy();
			else
				dump->create(getFieldLength(i),(char*)getFieldValue(i));
			break;
			}
		}
	return OK;
	}
//	________________________________________________________________

//	-------------					Printing							----------
//	________________________________________________________________
void __Query::printField(DbSQLDA& s, int i, MemBuf& dest, char* format)	{
	out_sqlda = (SQLDA*)s;
	char* data = (char*)getFieldValue(i);
	switch (getFieldType(i))	{
		case SQL_BLOB 	:
		case SQL_LTEXT :	dest << "<BLOB>";										break;
		case SQL_SHORT	:	dest << StdFormat("%hd", *(short*)data);		break;
		case SQL_LONG	:	dest << StdFormat("%ld", *(long*)data);		break;
		case SQL_FLOAT	:	dest << StdFormat("%g", *(float*)data);		break;
		case SQL_DOUBLE:	dest << StdFormat("%g", *(double*)data);		break;
		case SQL_DATE 	:	{
			Date a;
			getArg(s, &a, Ft_Date_, i);
			dest << a.format(format);
			break;
			}
		default :	{
			data = getTextField(i);
			Dump swap(1 + strlen(data));
			CpgConvertString(data,Cpg_CP_DATABASE,swap.data,Cpg_CP_DEFAULT,0);
			dest << swap.data;
			}
		}
	}

void __Query::print(DbSQLDA& s, MemBuf& dest)	{
	out_sqlda = (SQLDA*)s;
	for (int i = 0; i < out_sqlda->sqld; i++)	{
		dest << StdFormat("%32.32s= ",getFieldName(i));
		printField(s,i,dest);
		dest << "\n";
		}
	}
//	________________________________________________________________


Result __Query::setUniArg(DbSQLDA& sqlda, int i, UniVar* uv)	{
	SQLVAR& var = sqlda[i];
	switch (var.sqltype)	{
		case SQL_SHORT :	*(short*)var.sqldata = *(short*)uv->value;		break;
		case SQL_LONG :	*(long*)var.sqldata 	= *(long*)uv->value;			break;
		case SQL_FLOAT :	*(float*)var.sqldata = *(float*)uv->value;		break;
		case SQL_DOUBLE :	*(double*)var.sqldata= *(double*)uv->value;		break;
		case SQL_FIXCHAR :
		case SQL_VARCHAR :	{
			int maxlen = sqlda.record.get(i)->length - sizeof(DbRecordField);
			int ln = strlen((char*)uv->value);
			if (ln > maxlen)
				return queryError(QUE_TRUNCATION,uv->value);
			strcpy(var.sqldata,(char*)uv->value);
			var.sqllen = ln;
			break;
			}
		default :
			return queryError(QUE_TYPE_NOT_SUPPORTED,var.sqltype);
		}
	return OK;
	}

Result __Query::getUniArg(DbSQLDA& sqlda, int i, UniVar* uv)	{
	SQLVAR& var = sqlda[i];
	uv->uniType = var.sqltype;
	switch (var.sqltype)	{
		case SQL_SHORT :
		case SQL_LONG :
		case SQL_FLOAT :
		case SQL_DOUBLE :
		case SQL_FIXCHAR :
		case SQL_VARCHAR :
			uv->value = var.sqldata;
			break;
		default :
			return queryError(QUE_TYPE_NOT_SUPPORTED,var.sqltype);
		}
	return OK;
	}
//	________________________________________________________________

Result __Query::play(Dump& data, char* queryText, ...)	{
	if (prepare(queryText) == FAIL)
		return FAIL;

	va_list  ap;
	va_start(ap, queryText);
	in.clear();
	fillSQLDA(in,ap);
	va_end(ap);
	in.synchronize();
	ExactString req = queryText;
	data << (Dump&)in.sqlda << (Dump&)in.record << (Dump&)req;
	return _execute(in);
	}

Result __Query::replay(Dump& data)	{
	in.clear();
	ExactString req;
	data >> (Dump&)req >> (Dump&)in.record >> (Dump&)in.sqlda;
	in.synchronize();

	if (prepare((char*)req) == FAIL)
		return FAIL;

/*
	SQLDA* s = (SQLDA*)in;
	LstList* blobs = &(Transaction::getActive()->getDb()->blobList);
	for (int i = 0; i < sqlda->sqld; i++)
		if ((sqlda->sqlvar[i].sqltype & ~1) == SQL_BLOB)	{
			BlobItem* item;
			for (LstIterForward(*blobs,item))
				if (item->oldId == *(DbBlobId*)sqlda->sqlvar[i].sqldata)	{
					*(DbBlobId*)sqlda->sqlvar[i].sqldata = item->newId;
					break;
					}
			}
	LstFlush(*blobs,BlobItem);
*/
	return _execute(in);
	}

